"""Code generation utilities"""
import os
import pprint
import re
import sys
import textwrap

import jsonschema

from .utils import (CustomPrettyPrinter, SchemaInfo, is_valid_identifier, indent_docstring, indent_arglist,
                    load_metaschema)
from importlib.util import module_from_spec, spec_from_loader


class CodeSnippet(object):
    """Object whose repr() is a string of code"""

    def __init__(self, code):
        self.code = code

    def __repr__(self):
        return self.code


def _get_args(info):
    """Return the list of args & kwds for building the __init__ function"""
    # TODO: - set additional properties correctly
    #       - handle patternProperties etc.
    required = set()
    kwds = set()
    invalid_kwds = set()

    # TODO: specialize for anyOf/oneOf?

    if info.is_allOf():
        # recursively call function on all children
        arginfo = [_get_args(child) for child in info.allOf]
        nonkeyword = all(args[0] for args in arginfo)
        required = set.union(set(), *(args[1] for args in arginfo))
        kwds = set.union(set(), *(args[2] for args in arginfo))
        invalid_kwds = set.union(set(), *(args[3] for args in arginfo))
        additional = all(args[4] for args in arginfo)
    elif info.is_empty() or info.is_compound():
        nonkeyword = True
        additional = True
    elif info.is_value():
        nonkeyword = True
        additional = False
    elif info.is_object():
        invalid_kwds = ({p for p in info.required if not is_valid_identifier(p)} |
                        {p for p in info.properties if not is_valid_identifier(p)})
        required = {p for p in info.required if is_valid_identifier(p)}
        kwds = {p for p in info.properties if is_valid_identifier(p)}
        kwds -= required
        nonkeyword = False
        additional = True
        # additional = info.additionalProperties or info.patternProperties
    else:
        raise ValueError("Schema object not understood")

    return (nonkeyword, required, kwds, invalid_kwds, additional)


class SchemaClassGenerator(object):
    """Class that defines methods for generating code from schemas

    Parameters
    ----------
    classname : string
        The name of the class to generate
    schema : dict
        The dictionary defining the schema class
    rootschema : dict (optional)
        The root schema for the class
    basename : string (default: "SchemaBase")
        The name of the base class to use in the class definition
    schemarepr : CodeSnippet or object, optional
        An object whose repr will be used in the place of the explicit schema.
        This can be useful, for example, when the generated code should reference
        a predefined schema object. The user must ensure that the schema within
        the evaluated code is identical to the schema used to generate the code.
    rootschemarepr : CodeSnippet or object, optional
        An object whose repr will be used in the place of the explicit root
        schema.
    """
    schema_class_template = textwrap.dedent('''
    class {classname}({basename}):
        """{docstring}"""
        _schema = {schema!r}
        _rootschema = {rootschema!r}
        _property_names = {property_names!r}

        {init_code}
    ''')

    init_template = textwrap.dedent("""
    def __init__({arglist}):
        super({classname}, self).__init__({super_arglist})
    """).lstrip()

    def _process_description(self, description):
        return description

    def __init__(self, classname, schema, rootschema=None,
                 basename='SchemaBase', schemarepr=None, rootschemarepr=None,
                 nodefault=()):
        self.classname = classname
        self.schema = schema
        self.rootschema = rootschema
        self.basename = basename
        self.schemarepr = schemarepr
        self.rootschemarepr = rootschemarepr
        self.nodefault = nodefault

    def schema_class(self):
        """Generate code for a schema class"""
        rootschema = self.rootschema if self.rootschema is not None else self.schema
        schemarepr = self.schemarepr if self.schemarepr is not None else self.schema
        rootschemarepr = self.rootschemarepr
        property_names = tuple(self.schema.get('properties', {}).keys())
        if not len(property_names) and self.schema.get('additionalProperties', True):
            property_names = None

        if rootschemarepr is None:
            if rootschema is self.schema:
                rootschemarepr = CodeSnippet('_schema')
            else:
                rootschemarepr = rootschema
        return self.schema_class_template.format(
                classname=self.classname,
                basename=self.basename,
                schema=schemarepr,
                rootschema=rootschemarepr,
                docstring=self.docstring(indent=4),
                init_code=self.init_code(indent=4),
                property_names=property_names
        )

    def docstring(self, indent=0):
        # TODO: add a general description at the top, derived from the schema.
        #       for example, a non-object definition should list valid type, enum
        #       values, etc.
        # TODO: use _get_args here for more information on allOf objects
        info = SchemaInfo(self.schema, self.rootschema)
        doc = ["{} schema wrapper".format(self.classname),
               '',
               info.medium_description]
        if info.description:
            doc += self._process_description(  # remove condition description
                    re.sub(r"\n\{\n(\n|.)*\n\}", '', info.description)).splitlines()

        if info.properties:
            nonkeyword, required, kwds, invalid_kwds, additional = _get_args(info)
            doc += ['',
                    'Attributes',
                    '----------',
                    '']
            for prop in sorted(required) + sorted(kwds) + sorted(invalid_kwds):
                if prop in info.properties:
                    propinfo = info.properties[prop]
                    doc += ["{} : {}".format(prop, propinfo.short_description),
                            "    {}".format(self._process_description(propinfo.description))]
        if len(doc) > 1:
            doc += ['']
        return indent_docstring(doc, indent_level=indent, width=100, lstrip=True)

    def init_code(self, indent=0):
        """Return code suitablde for the __init__ function of a Schema class"""
        info = SchemaInfo(self.schema, rootschema=self.rootschema)
        nonkeyword, required, kwds, invalid_kwds, additional = _get_args(info)

        nodefault = set(self.nodefault)
        required -= nodefault
        kwds -= nodefault

        args = ['self']
        super_args = []

        if nodefault:
            args.extend(sorted(nodefault))
        elif nonkeyword:
            args.append('*args')
            super_args.append('*args')

        args.extend('{}=Undefined'.format(p)
                    for p in sorted(required) + sorted(kwds))
        super_args.extend('{0}={0}'.format(p)
                          for p in sorted(nodefault) + sorted(required) + sorted(kwds))

        if additional:
            args.append('**kwds')
            super_args.append('**kwds')

        arg_indent_level = 9 + indent
        super_arg_indent_level = 23 + len(self.classname) + indent

        initfunc = self.init_template.format(classname=self.classname,
                                             arglist=indent_arglist(args, indent_level=arg_indent_level),
                                             super_arglist=indent_arglist(super_args, indent_level=super_arg_indent_level))
        if indent:
            initfunc = ('\n' + indent * ' ').join(initfunc.splitlines())
        return initfunc


class SchemaModuleGenerator(object):
    """Generate a Python module implementing the schema

    Parameters
    ----------
    schema : dict
        The root schema description
    root_name : string
        The name of the root class (default: 'Root')
    schemaperfect_import : string
        The import path for schemaperfect (default: 'schemaperfect')
    """

    schema_module_header = textwrap.dedent("""
    # Module generated by SchemaModuleGenerator

    from {schemaperfect} import SchemaBase, Undefined
    """)

    def __init__(self, schema, root_name='Root', schemaperfect_import='schemaperfect'):
        self.schema = schema
        self.root_name = root_name
        self.schemaperfect_import = schemaperfect_import
        self._validate()

    def _validate(self):
        metaschema = load_metaschema()
        jsonschema.validate(self.schema, metaschema)

    def module_code(self):
        """Generate a Python module implementing the schema"""
        definitions = self.schema.get('definitions', {})

        if self.root_name in definitions:
            raise ValueError(f"root_name='{self.root_name}' exists in definitions; "
                             "please choose a different name")

        code = ['"""Module generated by SchemaModuleGenerator"""',
                f"from {self.schemaperfect_import} import SchemaBase, Undefined"]

        pretty_printer_kwargs = dict(width=140, compact=False, indent=4)
        if sys.version_info.major == 3 and sys.version_info.minor >= 8:
            pretty_printer_kwargs['sort_dicts'] = False

        pretty_printer = CustomPrettyPrinter(**pretty_printer_kwargs)
        schemarepr = textwrap.indent(pretty_printer.pformat(object=self.schema), 4 * ' ').lstrip()
        root = SchemaClassGenerator(self.root_name, self.schema,
                                    schemarepr=CodeSnippet(schemarepr), )
        code.append(root.schema_class())

        for name, subschema in definitions.items():
            schemarepr = f"{{'$ref': '#/definitions/{name}'}}"
            rootschemarepr = f'{self.root_name}._schema'
            gen = SchemaClassGenerator(classname=name,
                                       schema=subschema,
                                       rootschema=self.schema,
                                       schemarepr=CodeSnippet(schemarepr),
                                       rootschemarepr=CodeSnippet(rootschemarepr))
            code.append(gen.schema_class())

        return '\n\n'.join(code)

    def write_module(self, modulename):
        """Write the schema module to the given filename
        
        Parameters
        ----------
        modulename : string or Path
            the path to the module (should end with a .py extension)

        Returns
        -------
        modulepath : string
            the full absolute path to the written module
        """
        modulename = os.fspath(modulename)  # support pathlib.Path & others
        code = self.module_code()
        with open(modulename, 'w') as f:
            f.write(code)
        return os.path.abspath(modulename)

    def import_as(self, modulename, add_to_sys_modules=True):
        """Import wrapper as a dynamically-generated module.

        Parameters
        ----------
        modulename : string
            a valid Python module name.
        add_to_sys_modules : boolean
            if True (default) then add the modulename to sys.modules to allow
            accessing the module contents via standard import statements.

        Returns
        -------
        module :
            the dynamically-created module.
        """
        module = module_from_spec(spec_from_loader(modulename, loader=None))
        exec(self.module_code(), module.__dict__)
        if add_to_sys_modules:
            sys.modules[modulename] = module
        return module



"""
schemaperfect: tools for generating Python APIs from JSON schemas
"""
from .schemaperfect import SchemaBase, Undefined, SchemaValidationError
from .decorator import schemaclass
from .utils import SchemaInfo
from .codegen import SchemaModuleGenerator
from .version import version as __version__


__all__ = (
    "SchemaBase",
    "Undefined",
    "schemaclass",
    "SchemaInfo",
    "SchemaModuleGenerator",
    "SchemaValidationError"
)



import pytest
from schemaperfect import SchemaBase, SchemaModuleGenerator


@pytest.fixture
def schema():
    return {
        'definitions': {
            'Person': {
                'properties': {
                    'name': {'type': 'string'},
                    'age': {'type': 'integer'},
                }
            }
        },
        'properties': {
            'family_name': {
                'type': 'string'
            },
            'people': {
                'type': 'array',
                'items': {'$ref': '#/definitions/Person'}
            }
        },
        'required': ['family_name']
    }


def test_module_code(schema):
    gen = SchemaModuleGenerator(schema, root_name='Family')
    code = gen.module_code()

    namespace = {}
    exec(code, namespace)

    Family = namespace['Family']
    Person = namespace['Person']

    assert issubclass(Family, SchemaBase)
    assert issubclass(Person, SchemaBase)
    assert Family._property_names == ('family_name', 'people')

    family = Family(family_name='Smith', people=[Person(name='Alice', age=25), Person(name='Bob', age=26)])
    dct = family.to_dict()
    assert dct == {'family_name': 'Smith', 'people': [{'name': 'Alice', 'age': 25}, {'name': 'Bob', 'age': 26}]}
    family2 = Family.from_dict(dct)
    assert family2.to_dict() == dct


# noinspection PyUnresolvedReferences
def test_dynamic_module(schema):
    gen = SchemaModuleGenerator(schema, root_name='Family')
    testmod = gen.import_as('testmod')
    assert issubclass(testmod.Family, SchemaBase)
    assert issubclass(testmod.Person, SchemaBase)

    from testmod import Family, Person
    assert issubclass(Family, SchemaBase)
    assert issubclass(Person, SchemaBase)

    family = Family(family_name='Smith', people=[Person(name='Alice', age=25), Person(name='Bob', age=26)])
    dct = family.to_dict()
    assert dct == {'family_name': 'Smith', 'people': [{'name': 'Alice', 'age': 25}, {'name': 'Bob', 'age': 26}]}
    family2 = Family.from_dict(dct)
    assert family2.to_dict() == dct

    # test extended class with properties
    class MyFamily(Family):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.dependants = 0

        @property
        def has_pet(self):
            return False

    family3 = MyFamily.from_dict(dct)
    assert family3.dependants == 0
    family3.dependants = 1
    assert family3.dependants == 1
    assert not family3.has_pet
    assert family3.to_dict() == dct



import jsonschema
import pytest

from ..schemaperfect import (UndefinedType, SchemaBase, Undefined, _FromDict,
                        SchemaValidationError)

# Make tests inherit from _TestSchema, so that when we test from_dict it won't
# try to use SchemaBase objects defined elsewhere as wrappers.

class _TestSchema(SchemaBase):
    @classmethod
    def _default_wrapper_classes(cls):
        return _TestSchema.__subclasses__()


class MySchema(_TestSchema):
    _schema = {
        'definitions': {
            'StringMapping': {'type': 'object', 'additionalProperties': {'type': 'string'}},
            'StringArray': {'type': 'array', 'items': {'type': 'string'}}
        },
        'properties': {
            'a': {'$ref': '#/definitions/StringMapping'},
            'a2': {'type': 'object', 'additionalProperties': {'type': 'number'}},
            'b': {'$ref': '#/definitions/StringArray'},
            'b2': {'type': 'array', 'items': {'type': 'number'}},
            'c': {'type': ['string', 'number']},
            'd': {'anyOf': [{'$ref': '#/definitions/StringMapping'},
                            {'$ref': '#/definitions/StringArray'}]}
        }
    }


class StringMapping(_TestSchema):
    _schema = {'$ref': '#/definitions/StringMapping'}
    _rootschema = MySchema._schema


class StringArray(_TestSchema):
    _schema = {'$ref': '#/definitions/StringArray'}
    _rootschema = MySchema._schema


class Derived(_TestSchema):
    _schema = {
        'definitions': {
            'Foo': {
                'type': 'object',
                'properties': {
                    'd': {'type': 'string'}
                }
            },
            'Bar': {'type': 'string', 'enum': ['A', 'B']}
        },
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'a': {'type': 'integer'},
            'b': {'type': 'string'},
            'c': {"$ref": "#/definitions/Foo"}
        }
    }


class Foo(_TestSchema):
    _schema = {"$ref": "#/definitions/Foo"}
    _rootschema = Derived._schema


class Bar(_TestSchema):
    _schema = {"$ref": "#/definitions/Bar"}
    _rootschema = Derived._schema


class SimpleUnion(_TestSchema):
    _schema = {'anyOf' : [{'type': 'integer'}, {'type': 'string'}]}


class DefinitionUnion(_TestSchema):
    _schema = {
        "anyOf": [
            {"$ref": "#/definitions/Foo"},
            {"$ref": "#/definitions/Bar"}
        ]
    }
    _rootschema = Derived._schema


class SimpleArray(_TestSchema):
    _schema = {
        'type': 'array',
        'items': {
            'anyOf' : [{'type': 'integer'}, {'type': 'string'}]
        }
    }


class InvalidProperties(_TestSchema):
    _schema = {
        'type': 'object',
        'properties': {
            'for': {},
            'as': {},
            'vega-lite': {},
            '$schema': {}
        }
    }


def test_construct_multifaceted_schema():
    dct = {'a': {'foo': 'bar'}, 'a2': {'foo': 42},
       'b': ['a', 'b', 'c'], 'b2': [1, 2, 3], 'c': 42,
       'd': ['x', 'y', 'z']}

    myschema = MySchema.from_dict(dct)
    assert myschema.to_dict() == dct

    myschema2 = MySchema(**dct)
    assert myschema2.to_dict() == dct

    assert isinstance(myschema.a, StringMapping)
    assert isinstance(myschema.a2, dict)
    assert isinstance(myschema.b, StringArray)
    assert isinstance(myschema.b2, list)
    assert isinstance(myschema.d, StringArray)


def test_schema_cases():
    assert Derived(a=4, b='yo').to_dict() == {'a': 4, 'b': 'yo'}
    assert Derived(a=4, c={'d': 'hey'}).to_dict() == {'a': 4, 'c': {'d': 'hey'}}
    assert Derived(a=4, b='5', c=Foo(d='val')).to_dict() == {'a': 4, 'b': '5', 'c': {'d': 'val'}}
    assert Foo(d='hello', f=4).to_dict() == {'d': 'hello', 'f': 4}

    assert Derived().to_dict() == {}
    assert Foo().to_dict() == {}

    with pytest.raises(jsonschema.ValidationError):
        # a needs to be an integer
        Derived(a='yo').to_dict()

    with pytest.raises(jsonschema.ValidationError):
        # Foo.d needs to be a string
        Derived(c=Foo(4)).to_dict()

    with pytest.raises(jsonschema.ValidationError):
        # no additional properties allowed
        Derived(foo='bar').to_dict()


def test_round_trip():
    D = {'a': 4, 'b': 'yo'}
    assert Derived.from_dict(D).to_dict() == D

    D = {'a': 4, 'c': {'d': 'hey'}}
    assert Derived.from_dict(D).to_dict() == D

    D = {'a': 4, 'b': '5', 'c': {'d': 'val'}}
    assert Derived.from_dict(D).to_dict() == D

    D = {'d': 'hello', 'f': 4}
    assert Foo.from_dict(D).to_dict() == D


def test_from_dict():
    D = {'a': 4, 'b': '5', 'c': {'d': 'val'}}
    obj = Derived.from_dict(D)
    assert obj.a == 4
    assert obj.b == '5'
    assert isinstance(obj.c, Foo)


def test_simple_type():
    assert SimpleUnion(4).to_dict() == 4


def test_simple_array():
    assert SimpleArray([4, 5, 'six']).to_dict() == [4, 5, 'six']
    assert SimpleArray.from_dict(list('abc')).to_dict() == list('abc')


def test_definition_union():
    obj = DefinitionUnion.from_dict("A")
    assert isinstance(obj, Bar)
    assert obj.to_dict() == "A"

    obj = DefinitionUnion.from_dict("B")
    assert isinstance(obj, Bar)
    assert obj.to_dict() == "B"

    obj = DefinitionUnion.from_dict({'d': 'yo'})
    assert isinstance(obj, Foo)
    assert obj.to_dict() == {'d': 'yo'}


def test_invalid_properties():
    dct = {'for': 2, 'as': 3, 'vega-lite': 4, '$schema': 5}
    invalid = InvalidProperties.from_dict(dct)
    assert invalid['for'] == 2
    assert invalid['as'] == 3
    assert invalid['vega-lite'] == 4
    assert invalid['$schema'] == 5
    assert invalid.to_dict() == dct


def test_undefined_singleton():
    assert Undefined is UndefinedType()


def test_copy():
    dct = {'a': {'foo': 'bar'}, 'a2': {'foo': 42},
       'b': ['a', 'b', 'c'], 'b2': [1, 2, 3], 'c': 42,
       'd': ['x', 'y', 'z']}

    myschema = MySchema.from_dict(dct)

    # Make sure copy is deep
    copy = myschema.copy(deep=True)
    copy['a']['foo'] = 'new value'
    copy['b'] = ['A', 'B', 'C']
    copy['c'] = 164
    assert myschema.to_dict() == dct

    # If we exclude a value, changing the copy changes the original
    copy = myschema.copy(deep=True, exclude=['a'])
    copy['a']['foo'] = 'new value'
    copy['b'] = ['A', 'B', 'C']
    copy['c'] = 164
    mydct = myschema.to_dict()
    assert mydct['a']['foo'] == 'new value'
    assert mydct['b'][0] == dct['b'][0]
    assert mydct['c'] == dct['c']

    # If copy is not deep, then changing copy below top level changes original
    copy = myschema.copy(deep=False)
    copy['a']['foo'] = 'baz'
    copy['b'] = ['A', 'B', 'C']
    copy['c'] = 164
    mydct = myschema.to_dict()
    assert mydct['a']['foo'] == 'baz'
    assert mydct['b'] == dct['b']
    assert mydct['c'] == dct['c']


def test_attribute_error():
    m = MySchema()
    with pytest.raises(AttributeError) as err:
        m.invalid_attribute
    assert str(err.value) == ("'MySchema' object has no attribute "
                              "'invalid_attribute'")


def test_to_from_json():
    dct = {'a': {'foo': 'bar'}, 'a2': {'foo': 42},
           'b': ['a', 'b', 'c'], 'b2': [1, 2, 3], 'c': 42,
           'd': ['x', 'y', 'z']}
    json_str = MySchema.from_dict(dct).to_json()
    new_dct = MySchema.from_json(json_str).to_dict()

    assert new_dct == dct


def test_class_with_no_schema():
    class BadSchema(SchemaBase):
        pass

    with pytest.raises(ValueError) as err:
        BadSchema(4)
    assert str(err.value).startswith("Cannot instantiate object")


@pytest.mark.parametrize('use_json', [True, False])
def test_hash_schema(use_json):
    classes = _TestSchema._default_wrapper_classes()

    for cls in classes:
        hsh1 = _FromDict.hash_schema(cls._schema, use_json=use_json)
        hsh2 = _FromDict.hash_schema(cls._schema, use_json=use_json)
        assert hsh1 == hsh2
        assert hash(hsh1) == hash(hsh2)


def test_schema_validation_error():
    try:
        MySchema(a={'foo': 4})
        the_err = None
    except jsonschema.ValidationError as err:
        the_err = err

    assert isinstance(the_err, SchemaValidationError)
    message = str(the_err)

    assert message.startswith('Invalid specification')
    assert 'test_schemaperfect.MySchema->a' in message
    assert "validating {!r}".format(the_err.validator) in message
    assert the_err.message in message



import sys
from textwrap import dedent

import pytest

from ..utils import CustomPrettyPrinter, get_valid_identifier, load_metaschema
from ..schemaperfect import _FromDict, set_metaschema_version


@pytest.fixture
def refschema():
    return {
        '$ref': '#/definitions/Foo',
        'definitions': {
            'Foo': {'$ref': '#/definitions/Bar'},
            'Bar': {'$ref': '#/definitions/Baz'},
            'Baz': {'type': 'string'}
        }
    }


def test_get_valid_identifier():
    assert get_valid_identifier('$schema') == 'schema'
    assert get_valid_identifier('$ref') == 'ref'
    assert get_valid_identifier('foo-bar') == 'foobar'
    assert get_valid_identifier('$as') == 'as_'
    assert get_valid_identifier('for') == 'for_'
    assert get_valid_identifier('--') == '_'


@pytest.mark.parametrize('use_json', [True, False])
def test_hash_schema(refschema, use_json):
    copy = refschema.copy()
    copy['description'] = "A schema"
    copy['title'] = "Schema to test"
    assert _FromDict.hash_schema(refschema) == _FromDict.hash_schema(copy)


@pytest.mark.parametrize('draft_no', ['07', '06', '04', '03'])
def test_metaschema_version(draft_no):
    set_metaschema_version('draft' + str(int(draft_no)))
    metaschema = load_metaschema()
    id_key = "$id" if int(draft_no) >= 6 else "id"
    assert metaschema[id_key].split('//')[1].startswith("json-schema.org/draft-{0}".format(draft_no))

def test_custom_pretty_printer(refschema):
    pretty_printer_kwargs = dict(width=80, compact=False, indent=4)
    if sys.version_info.major == 3 and sys.version_info.minor >= 8:
        pretty_printer_kwargs['sort_dicts'] = False
    pretty_printer = CustomPrettyPrinter(**pretty_printer_kwargs)
    print(pretty_printer.pformat(object=refschema))
    assert pretty_printer.pformat(object=refschema) == dedent("""
    {
        '$ref': '#/definitions/Foo',
        'definitions': {
            'Bar': {'$ref': '#/definitions/Baz'},
            'Baz': {'type': 'string'},
            'Foo': {'$ref': '#/definitions/Bar'}
        }
    }
    """).strip('\n')




import warnings
from . import codegen, SchemaBase, Undefined


def schemaclass(*args, init_func=True, docstring=True, property_map=True):
    """A decorator to add boilerplate to a schema class

    This will read the _json_schema attribute of a SchemaBase class, and add
    one or all of three attributes/methods, based on the schema:

    - An __init__ function
    - a __doc__ docstring

    In all cases, if the attribute/method is explicitly defined in the class
    it will not be overwritten.

    A simple invocation adds all three attributes/methods:

        @schemaclass
        class MySchema(SchemaBase):
            _schema = {...}

    Optionally, you can invoke schemaclass with arguments to turn off
    some of the added behaviors:

        @schemaclass(init_func=True, docstring=False)
        class MySchema(SchemaBase):
            _schema = {...}
    """
    def _decorator(cls, init_func=init_func, docstring=docstring):
        if not (isinstance(cls, type) and issubclass(cls, SchemaBase)):
            warnings.warn("class is not an instance of SchemaBase.")

        name = cls.__name__
        gen = codegen.SchemaClassGenerator(name, schema=cls._schema,
                                      rootschema=cls._rootschema)

        if init_func and '__init__' not in cls.__dict__:
            init_code = gen.init_code()
            globals_ = {name: cls, 'Undefined': Undefined}
            locals_ = {}
            exec(init_code, globals_, locals_)
            setattr(cls, '__init__', locals_['__init__'])

        if docstring and not cls.__doc__:
            setattr(cls, '__doc__', gen.docstring())
        return cls

    if len(args) == 0:
        return _decorator
    elif len(args) == 1:
        return _decorator(args[0])
    else:
        raise ValueError("optional arguments to schemaclass must be "
                         "passed by keyword")




# This file is automatically generated by the setup.py script
long_description = """
JSONSchema API Generator
========================
The schemaperfect_ package provides tools for auto-generation of Python
APIs from JSONSchema_ specifications.

See more information in the README_.

.. _README: https://github.com/jwilson8767/schemaperfect/blob/master/README.md
.. _JSONSchema: http://json-schema.org/
.. _schemaperfect: https://github.com/jwilson8767/schemaperfect
"""
short_version = '0.5.0'
version = '0.5.0'
full_version = '0.5.0'
git_revision = 'ffc2ca9ded10bc31651e6867d630b0d3d5daf8d9'
release = True
if not release:
    version = full_version



#!/usr/bin/env python
"""
Run `./setup.py test` to execute tests or `python -m pytest schemaperfect --doctest-modules`
Run `./setup.py build` to build
Dev install via `pip install -e .`
"""


import os
import subprocess
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

# Version info
MAJOR = 0
MINOR = 5
MICRO = 0
ISRELEASED = True
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)

CLASSIFIERS = ["Development Status :: 3 - Alpha",
               "Environment :: Console",
               "Intended Audience :: Science/Research",
               "License :: OSI Approved :: BSD License",
               "Operating System :: OS Independent",
               "Programming Language :: Python :: 3.6",
               "Programming Language :: Python :: 3.7",
               "Topic :: Scientific/Engineering"]

# Description should be a one-liner:
DESCRIPTION = "schemaperfect: generate Python APIs from JSONSchema specifications"

# Long description will go up on the pypi page
LONG_DESCRIPTION = """
JSONSchema API Generator
========================
The schemaperfect_ package provides tools for auto-generation of Python
APIs from JSONSchema_ specifications.

See more information in the README_.

.. _README: https://github.com/jwilson8767/schemaperfect/blob/master/README.md
.. _JSONSchema: http://json-schema.org/
.. _schemaperfect: https://github.com/jwilson8767/schemaperfect
"""

###############################################################################
# The following versioning code is adapted from the scipy package (BSD license)
# http://github.com/scipy/scipy
# (C) 2003-2017 SciPy Developers

def git_version():
    """Return the git revision as a string"""
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        GIT_REVISION = out.strip().decode('ascii')
    except OSError:
        GIT_REVISION = "Unknown"

    return GIT_REVISION

def get_version_info():
    """
    Adding the git rev number needs to be done inside
    write_version_py(), otherwise the import of version messes
    up the build under Python 3.
    """
    FULLVERSION = VERSION
    if os.path.exists('.git'):
        GIT_REVISION = git_version()
    elif os.path.exists('schemaperfect/version.py'):
        # must be a source distribution, use existing version file
        # load it as a separate module to not load schemaperfect/__init__.py
        import imp
        version = imp.load_source('schemaperfect.version',
                                  'schemaperfect/version.py')
        GIT_REVISION = version.git_revision
    else:
        GIT_REVISION = "Unknown"

    if not ISRELEASED:
        FULLVERSION += '.dev0+' + GIT_REVISION[:7]

    return FULLVERSION, GIT_REVISION


def write_version_py(filename='schemaperfect/version.py'):
    cnt = '''
# This file is automatically generated by the setup.py script
long_description = """{long_description}"""
short_version = '{version}'
version = '{version}'
full_version = '{full_version}'
git_revision = '{git_revision}'
release = {isrelease}
if not release:
    version = full_version
'''
    FULLVERSION, GIT_REVISION = get_version_info()
    with open(filename, 'w') as f:
        f.write(cnt.format(long_description=LONG_DESCRIPTION,
                           version=VERSION,
                           full_version=FULLVERSION,
                           git_revision=GIT_REVISION,
                           isrelease=str(ISRELEASED)))

# End of code adapted from SciPy
###############################################################################

def setup_package():
    write_version_py()

    metadata = dict(
        name="schemaperfect",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        maintainer="Joshua Wilson",
        maintainer_email="jwilson8767@gmail.com",
        url="http://github.com/jwilson8767/schemaperfect",
        include_package_data=True,
        license="BSD",
        author="Joshua Wilson",
        author_email="jwilson8767@gmail.com",
        platforms="OS Independent",
        package_data={},
        install_requires=["jsonschema"],
        python_requires='>3.6',
        tests_require=["pytest"],
        cmdclass={
            'test': PyTest,
        },
    )

    metadata['version'] = get_version_info()[0]
    metadata['packages'] = find_packages()

    setup(**metadata)


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['schemaperfect', '--doctest-modules']
        self.test_suite = True

    def run_tests(self):

        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


if __name__ == '__main__':
    setup_package()



import inspect

from .. import SchemaBase, Undefined, schemaclass

@schemaclass
class MySchema(SchemaBase):
    _schema = {
        'definitions': {
            'StringMapping': {'type': 'object', 'additionalProperties': {'type': 'string'}},
            'StringArray': {'type': 'array', 'items': {'type': 'string'}}
        },
        'properties': {
            'a': {'$ref': '#/definitions/StringMapping'},
            'a2': {'type': 'object', 'additionalProperties': {'type': 'number'}},
            'b': {'$ref': '#/definitions/StringArray'},
            'b2': {'type': 'array', 'items': {'type': 'number'}},
            'c': {'type': ['string', 'number']},
            'd': {'anyOf': [{'$ref': '#/definitions/StringMapping'},
                            {'$ref': '#/definitions/StringArray'}]}
        },
        'required': ['a', 'b']
    }


@schemaclass
class StringArray(SchemaBase):
    _schema = MySchema._schema['definitions']['StringArray']
    _rootschema = MySchema._schema


def test_myschema_decorator():
    myschema = MySchema({'foo': 'bar'}, ['foo', 'bar'])
    assert myschema.to_dict() == {'a': {'foo': 'bar'}, 'b': ['foo', 'bar']}

    myschema = MySchema.from_dict({'a': {'foo': 'bar'}, 'b': ['foo', 'bar']})
    assert myschema.to_dict() == {'a': {'foo': 'bar'}, 'b': ['foo', 'bar']}

    assert MySchema.__doc__.startswith('MySchema schema wrapper')
    argspec = inspect.getfullargspec(MySchema.__init__)
    assert argspec.args == ['self', 'a' ,'b', 'a2', 'b2', 'c', 'd']
    assert argspec.defaults == 6 * (Undefined,)
    assert argspec.varargs is None
    assert argspec.varkw == 'kwds'


def test_stringarray_decorator():
    arr = StringArray(['a', 'b', 'c'])
    assert arr.to_dict() == ['a', 'b', 'c']

    arr = StringArray.from_dict(['a', 'b', 'c'])
    assert arr.to_dict() == ['a', 'b', 'c']

    assert arr.__doc__.startswith('StringArray schema wrapper')
    argspec = inspect.getfullargspec(StringArray.__init__)
    assert argspec.args == ['self']
    assert argspec.varargs == 'args'
    assert argspec.varkw is None



"""Utilities for working with schemas"""

import json
import keyword
import pkgutil
import pprint
import re
import textwrap
from typing import IO, Optional

import jsonschema
import typing

EXCLUDE_KEYS = ('definitions', 'title', 'description', '$schema', 'id')


def load_metaschema():
    """Load the metaschema / jsonschema spec to validate the schema that will be used for code generation."""
    from schemaperfect.schemaperfect import get_metaschema_version
    schema = pkgutil.get_data("jsonschema", "schemas/{0}.json".format(get_metaschema_version()))
    schema = schema.decode('utf-8')
    return json.loads(schema)


def resolve_references(schema, root=None):
    """Resolve References within a JSON schema"""
    resolver = jsonschema.RefResolver.from_schema(root or schema)
    while '$ref' in schema:
        with resolver.resolving(schema['$ref']) as resolved:
            schema = resolved
    return schema


def get_valid_identifier(prop, replacement_character='', allow_unicode=False):
    """Given a string property, generate a valid Python identifier

    Parameters
    ----------
    replacement_character: string, default ''
        The character to replace invalid characters with.
    allow_unicode: boolean, default False
        If True, then allow Python 3-style unicode identifiers.

    Examples
    --------
    >>> get_valid_identifier('my-var')
    'myvar'

    >>> get_valid_identifier('if')
    'if_'

    >>> get_valid_identifier('$schema', '_')
    '_schema'

    >>> get_valid_identifier('$*#$')
    '_'
    """
    # First substitute-out all non-valid characters.
    flags = re.UNICODE if allow_unicode else re.ASCII
    valid = re.sub(r'\W', replacement_character, prop, flags=flags)

    # If nothing is left, use just an underscore
    if not valid:
        valid = '_'

    # first character must be a non-digit. Prefix with an underscore
    # if needed
    if re.match(r'^[d\W]', valid):
        valid = '_' + valid

    # if the result is a reserved keyword, then add an underscore at the end
    if keyword.iskeyword(valid):
        valid += '_'
    return valid


def is_valid_identifier(var, allow_unicode=False):
    """Return true if var contains a valid Python identifier

    Parameters
    ----------
    val : string
        identifier to check
    allow_unicode : bool (default: False)
        if True, then allow Python 3 style unicode identifiers.
    """
    flags = re.UNICODE if allow_unicode else re.ASCII
    is_valid = re.match(r"^[^\d\W]\w*\Z", var, flags)
    return is_valid and not keyword.iskeyword(var)


class SchemaProperties(object):
    """A wrapper for properties within a schema"""

    def __init__(self, properties, schema, rootschema=None):
        self._properties = properties
        self._schema = schema
        self._rootschema = rootschema or schema

    def __bool__(self):
        return bool(self._properties)

    def __dir__(self):
        return list(self._properties.keys())

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            return super().__getattr__(attr)

    def __getitem__(self, attr):
        dct = self._properties[attr]
        if 'definitions' in self._schema and 'definitions' not in dct:
            dct = dict(definitions=self._schema['definitions'], **dct)
        return SchemaInfo(dct, self._rootschema)

    def __iter__(self):
        return iter(self._properties)

    def items(self):
        return ((key, self[key]) for key in self)

    def keys(self):
        return self._properties.keys()

    def values(self):
        return (self[key] for key in self)


class SchemaInfo(object):
    """A wrapper for inspecting a JSON schema"""

    def __init__(self, schema, rootschema=None, validate=False):
        if hasattr(schema, '_schema'):
            if hasattr(schema, '_rootschema'):
                schema, rootschema = schema._schema, schema._rootschema
            else:
                schema, rootschema = schema._schema, schema._schema
        elif not rootschema:
            rootschema = schema
        if validate:
            metaschema = load_metaschema()
            jsonschema.validate(schema, metaschema)
            jsonschema.validate(rootschema, metaschema)
        self.raw_schema = schema
        self.rootschema = rootschema
        self.schema = resolve_references(schema, rootschema)

    def child(self, schema):
        return self.__class__(schema, rootschema=self.rootschema)

    def __repr__(self):
        keys = []
        for key in sorted(self.schema.keys()):
            val = self.schema[key]
            rval = repr(val).replace('\n', '')
            if len(rval) > 30:
                rval = rval[:30] + '...'
            if key == 'definitions':
                rval = "{...}"
            elif key == 'properties':
                rval = '{\n    ' + '\n    '.join(sorted(map(repr, val))) + '\n  }'
            keys.append('"{}": {}'.format(key, rval))
        return "SchemaInfo({\n  " + '\n  '.join(keys) + "\n})"

    @property
    def title(self):
        if self.is_reference():
            return get_valid_identifier(self.refname)
        else:
            return ''

    @property
    def short_description(self):
        if self.title:
            # use RST syntax for generated sphinx docs
            return ":class:`{}`".format(self.title)
        else:
            return self.medium_description

    @property
    def medium_description(self):
        _simple_types = {'string': 'string',
                         'number': 'float',
                         'integer': 'integer',
                         'object': 'mapping',
                         'boolean': 'boolean',
                         'array': 'list',
                         'null': 'None'}
        if self.is_empty():
            return 'any object'
        elif self.is_enum():
            return 'enum({})'.format(', '.join(map(repr, self.enum)))
        elif self.is_anyOf():
            return 'anyOf({})'.format(', '.join(s.short_description
                                                for s in self.anyOf))
        elif self.is_oneOf():
            return 'oneOf({})'.format(', '.join(s.short_description
                                                for s in self.oneOf))
        elif self.is_allOf():
            return 'allOf({})'.format(', '.join(s.short_description
                                                for s in self.allOf))
        elif self.is_not():
            return 'not {}'.format(self.not_.short_description)
        elif isinstance(self.type, typing.Sequence) and not isinstance(self.type, str):
            options = []
            subschema = SchemaInfo(dict(**self.schema))
            for typ_ in self.type:
                subschema.schema['type'] = typ_
                options.append(subschema.short_description)
            return "anyOf({})".format(', '.join(options))
        elif self.is_object():
            return "Mapping(required=[{}])".format(', '.join(self.required))
        elif self.is_array():
            return "List({})".format(self.child(self.items).short_description)
        elif self.type in _simple_types:
            return _simple_types[self.type]
        elif not self.type:
            import warnings
            warnings.warn("no short_description for schema\n{}"
                          "".format(self.schema))
            return 'any'

    @property
    def long_description(self):
        # TODO
        return 'Long description including arguments and their types'

    @property
    def properties(self):
        return SchemaProperties(self.schema.get('properties', {}),
                                self.schema, self.rootschema)

    @property
    def definitions(self):
        return SchemaProperties(self.schema.get('definitions', {}),
                                self.schema, self.rootschema)

    @property
    def required(self):
        return self.schema.get('required', [])

    @property
    def patternProperties(self):
        return self.schema.get('patternProperties', {})

    @property
    def additionalProperties(self):
        return self.schema.get('additionalProperties', True)

    @property
    def type(self):
        return self.schema.get('type', None)

    @property
    def anyOf(self):
        return [self.child(s) for s in self.schema.get('anyOf', [])]

    @property
    def oneOf(self):
        return [self.child(s) for s in self.schema.get('oneOf', [])]

    @property
    def allOf(self):
        return [self.child(s) for s in self.schema.get('allOf', [])]

    @property
    def not_(self):
        return self.child(self.schema.get('not_', {}))

    @property
    def items(self):
        return self.schema.get('items', {})

    @property
    def enum(self):
        return self.schema.get('enum', [])

    @property
    def refname(self):
        return self.raw_schema.get('$ref', '#/').split('/')[-1]

    @property
    def ref(self):
        return self.raw_schema.get('$ref', None)

    @property
    def description(self):
        return self.raw_schema.get('description',
                                   self.schema.get('description', ''))

    def is_reference(self):
        return '$ref' in self.raw_schema

    def is_enum(self):
        return 'enum' in self.schema

    def is_empty(self):
        return set(self.schema.keys()) - set(EXCLUDE_KEYS) == {}

    def is_compound(self):
        return any(key in self.schema for key in ['anyOf', 'allOf', 'oneOf'])

    def is_anyOf(self):
        return 'anyOf' in self.schema

    def is_allOf(self):
        return 'allOf' in self.schema

    def is_oneOf(self):
        return 'oneOf' in self.schema

    def is_not(self):
        return 'not' in self.schema

    def is_object(self):
        if self.type == 'object':
            return True
        elif self.type is not None:
            return False
        elif self.properties or self.required or self.patternProperties or self.additionalProperties:
            return True
        else:
            raise ValueError("Unclear whether schema.is_object() is True")

    def is_value(self):
        return not self.is_object()

    def is_array(self):
        return (self.type == 'array')

    def schema_type(self):
        if self.is_empty():
            return 'empty'
        elif self.is_compound():
            for key in ['anyOf', 'oneOf', 'allOf']:
                if key in self.schema:
                    return key
        elif self.is_object():
            return 'object'
        elif self.is_array():
            return 'array'
        elif self.is_value():
            return 'value'
        else:
            raise ValueError("Unknown type with keys {}".format(self.schema))

    def property_name_map(self):
        """
        Return a mapping of schema property names to valid Python attribute names

        Only properties which are not valid Python identifiers will be included in
        the dictionary.
        """
        pairs = [(prop, get_valid_identifier(prop)) for prop in self.properties]
        return {prop: val for prop, val in pairs if prop != val}


def indent_arglist(args, indent_level, width=100, lstrip=True):
    """Indent an argument list for use in generated code"""
    wrapper = textwrap.TextWrapper(width=width,
                                   initial_indent=indent_level * ' ',
                                   subsequent_indent=indent_level * ' ',
                                   break_long_words=False)
    wrapped = '\n'.join(wrapper.wrap(', '.join(args)))
    if lstrip:
        wrapped = wrapped.lstrip()
    return wrapped


def indent_docstring(lines, indent_level, width=100, lstrip=True):
    """Indent a docstring for use in generated code"""
    final_lines = []

    for i, line in enumerate(lines):
        stripped = line.lstrip()
        if stripped:
            leading_space = len(line) - len(stripped)
            indent = indent_level + leading_space
            wrapper = textwrap.TextWrapper(width=width - indent,
                                           initial_indent=indent * ' ',
                                           subsequent_indent=indent * ' ',
                                           break_long_words=False,
                                           break_on_hyphens=False,
                                           drop_whitespace=True)
            list_wrapper = textwrap.TextWrapper(width=width - indent,
                                                initial_indent=indent * ' ' + '* ',
                                                subsequent_indent=indent * ' ' + '  ',
                                                break_long_words=False,
                                                break_on_hyphens=False,
                                                drop_whitespace=True)
            for _line in stripped.split("\n"):
                if _line == '':
                    final_lines.append('')
                elif _line.startswith('* '):
                    final_lines.extend(list_wrapper.wrap(_line[2:]))
                else:
                    final_lines.extend(wrapper.wrap(_line.lstrip()))

        # If this is the last line, put in an indent
        elif i + 1 == len(lines):
            final_lines.append(indent_level * ' ')
        # If it's not the last line, this is a blank line that should not indent.
        else:
            final_lines.append('')
    # Remove any trailing whitespaces on the right side
    stripped_lines = []
    for i, _line in enumerate(final_lines):
        if i + 1 == len(final_lines):
            stripped_lines.append(_line)
        else:
            stripped_lines.append(_line.rstrip())
    # Join it all together
    wrapped = '\n'.join(stripped_lines)
    if lstrip:
        wrapped = wrapped.lstrip()
    return wrapped


class CustomPrettyPrinter(pprint.PrettyPrinter):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._dispatch[dict.__repr__] = CustomPrettyPrinter._pprint_dict

    def _format_dict_items(self, items, stream, indent, allowance, context, level):
        write = stream.write
        indent += self._indent_per_level
        delimnl = ',\n' + (indent * ' ')
        last_index = len(items) - 1
        for i, (key, ent) in enumerate(items):
            last = i == last_index
            rep = self._repr(key, context, level)
            write(rep)
            write(': ')
            self._format(ent, stream, indent,
                         allowance if last else 1,
                         context, level)
            if not last:
                write(delimnl)

    def _pprint_dict(self, object, stream, indent, allowance, context, level, *args):
        write = stream.write
        write('{\n' + (indent * ' '))
        if self._indent_per_level > 1:
            write(self._indent_per_level * ' ')
        length = len(object)
        if length:
            items = sorted(object.items(), key=pprint._safe_tuple)
            self._format_dict_items(items, stream, indent, allowance + 1,
                                    context, level)
        write('\n' + (indent * ' ') + '}')
